#include <stdint.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "sound.h.15bit-64"
//#include "sound.h.1bit"

// Timer0 Clock Select and Output Compare value
#define TIMER0_TARGET_RATE (1000000UL/SOUND_RATE)  // target rate in microseconds
#define TIMER0_MAX_COUNT    256UL

#if \
     ( ( TIMER0_MAX_COUNT * 1000UL *    1UL ) / ( F_CPU / 1000UL ) ) > \
     TIMER0_TARGET_RATE
#  define TIMER0_PRESCALER 1
#  define TIMER0_CS        ( 0 << CS02 ) | ( 0 << CS01 ) | ( 1 << CS00 )
#warning choosing prescaler=1 CS=CS00
#elif \
     ( ( TIMER0_MAX_COUNT * 1000UL *    8UL ) / ( F_CPU / 1000UL ) ) > \
     TIMER0_TARGET_RATE
#  define TIMER0_PRESCALER 8
#  define TIMER0_CS        ( 0 << CS02 ) | ( 1 << CS01 ) | ( 0 << CS00 )
#warning choosing prescaler=8 CS=CS01

#elif \
     ( ( TIMER0_MAX_COUNT * 1000UL *   64UL ) / ( F_CPU / 1000UL ) ) > \
     TIMER0_TARGET_RATE
#  define TIMER0_PRESCALER 64
#  define TIMER0_CS        ( 0 << CS02 ) | ( 1 << CS01 ) | ( 1 << CS00 )
#warning choosing prescaler=64 CS=CS01|CS00

#elif \
     ( ( TIMER0_MAX_COUNT * 1000UL *  256UL ) / ( F_CPU / 1000UL ) ) > \
     TIMER0_TARGET_RATE
#  define TIMER0_PRESCALER 256
#  define TIMER0_CS        ( 1 << CS02 ) | ( 0 << CS01 ) | ( 0 << CS00 )
#warning choosing prescaler=256 CS=CS02

#elif \
     ( ( TIMER0_MAX_COUNT * 1000UL * 1024UL ) / ( F_CPU / 1000UL ) ) > \
     TIMER0_TARGET_RATE
#  define TIMER0_PRESCALER 1024
#  define TIMER0_CS        ( 1 << CS02 ) | ( 0 << CS01 ) | ( 1 << CS00 )
#warning choosing prescaler=1024 CS=CS02|CS00

#else
#  error TIMER0_TARGET_RATE not attainable
#endif

#define TIMER0_OUTPUT_COMPARE \
     ( TIMER0_TARGET_RATE * ( F_CPU / 100UL ) ) / \
     ( TIMER0_PRESCALER * 10000UL )

#define DEBOUNCE_RATE      250UL  // per second = every 4msec
#define DEBOUNCE_TICKS  (SOUND_RATE / DEBOUNCE_RATE)
#define KEY_INPUT       PINA

#define KEY     PA0
#define LED     PA1
#define SND0    PA2
#define SND1    PA3

void debounce(void);

void t0_init( void )
{
    // timer 0 (8bit) generates sound playback clock
    // prescale as needed, CTC mode
    TCCR0 = (1<< WGM01) | TIMER0_CS;

    // generate compare0 interrupts at SOUND_RATE per second
    OCR0 = TIMER0_OUTPUT_COMPARE;

    // enable timer0 compare match interrupt
    TIMSK |=  (1 << OCIE0);

#if  SOUND_ENCODING == BTC_ONE_BIT
    DDRA = (1 << LED) | (1 << SND0);
#endif

#if  SOUND_ENCODING == BTC_ONEPOINTFIVE_BIT
    DDRA = (1 << LED) | (1 << SND0) | (1 << SND1);
#endif

    PORTA = (1 << LED); // on my eval board the LEDs are active low
}


volatile uint8_t playing;            // flag - currently playing the tune
volatile uint8_t cnt;                // bit counter in current 8-bit stream chunk
volatile uint8_t sample;             // chunk of stream
volatile uint16_t currentsample;     // index into stream (actually char sound[])

#if  SOUND_ENCODING == BTC_ONEPOINTFIVE_BIT
volatile uint8_t lastbit;
#endif

// play sound.

ISR (SIG_OUTPUT_COMPARE0)
{
    static uint16_t time_to_debounce;

    if (playing) {
        if (++cnt & 8) {
            if (currentsample < SOUND_BYTES) { // load next sample
                sample = pgm_read_byte(sound + currentsample);
                currentsample++;
            } else {
                playing = 0;            // done
                PORTA |= (1 << LED);    // turn off LED
                PORTA &= ~((1 << SND0) | (1 << SND1));
                goto DONE;
            }
            cnt = 0;
        }

#if  SOUND_ENCODING == BTC_ONE_BIT
        // toggle single pin according to current high bit
        if (sample & 0x80)
            PORTA |= (1 << SND0);
        else
            PORTA &= ~(1 << SND0);
#endif

#if  SOUND_ENCODING == BTC_ONEPOINTFIVE_BIT

        // if two consecutive bits are identical in the stream, both outputs are driven.
        if ((sample & 0x80) == lastbit) {

            DDRA |= (1 << SND1);         // enable output on SND1
            if (sample & 0x80)
                PORTA |= ((1 << SND0) | (1 << SND1));     // activate both
            else
                PORTA &= ~((1 << SND0) | (1 << SND1));    // deactivate both
        } else {
            // two consecutive bits differ.
            // only SND1 used, SND2 is left as input (floating)
            DDRA &= ~(1 << SND1);         // make pin floating
            if (sample & 0x80)              // set the other bit accordingly
                PORTA |= (1 << SND0);     // activate SND0
            else
                PORTA &= ~(1 << SND0);    // deactivate SND0
        }
        lastbit = sample & 0x80;            // remember for next iteration
#endif
        sample <<= 1;

        DONE: ;
    }
    // other periodic stuff
    if (time_to_debounce-- == 0) {
        time_to_debounce = DEBOUNCE_TICKS;
        debounce();
    }
}


// play a fixed tune from char sound[]
int playsound()
{
    if (playing)
        return playing; // we're already playing

    cli();
    playing = 1;
    currentsample = 0;
    sei();

    // trigger load of next sample
    cnt = 7;

#if  SOUND_ENCODING == BTC_ONEPOINTFIVE_BIT
    DDRA |= (1 << SND0);              // direction = out
    PORTA &= ~(1 << SND0);            // reset both pins
#endif

#if  SOUND_ENCODING == BTC_ONE_BIT
    DDRA |= (1 << SND0) | (1 << SND1);        // direction = out
    PORTA &= ~((1 << SND0) | (1 << SND1));    // reset both pins
#endif

    PORTA &= ~(1 << LED);    // turn on LED

    return 0;   //  playback started at next COMP0 interrupt
}


// Peter Danegger debounce routines

uint8_t key_state;                        // debounced and inverted key state:
                                          // bit = 1: key pressed
uint8_t key_press;                        // key press detect or repeat

void debounce(void)                     // call every 4msec or so
{
  static uint8_t ct0, ct1;
  uint8_t i;

  i = key_state ^ ~KEY_INPUT;           // key changed ?
  ct0 = ~( ct0 & i );                   // reset or count ct0
  ct1 = ct0 ^ (ct1 & i);                        // reset or count ct1
  i &= ct0 & ct1;                       // count until roll over ?
  key_state ^= i;                       // then toggle debounced state
  key_press |= key_state & i;           // 0->1: key press detect
}


uint8_t get_key_press( uint8_t key_mask )
{
  cli();
  key_mask &= key_press;                        // read key(s)
  key_press ^= key_mask;                        // clear key(s)
  sei();
  return key_mask;
}


int main( void )
{
    key_state = 0;
    key_press = 0;

    t0_init();
    sei();

    for(;;){                            // main loop
        if( get_key_press( 1<<KEY ) )
            playsound();
    }
}
